package spotai;

import java.util.*;
import static java.lang.Math.*;
import static spotai.BonusItems.*;

public class Hypermax {
  
  public static final int CLONES = 0;
  public static final int SKIPS = 1;  

  private final Random random = new Random();
  
  private int rows;
  private int columns;
  
  private Node[][] children;
  private Coordinate[][] coordinates;
  private Coordinate[][][][] moves;  
  private float[][] alphas;
  private boolean[][] topology;
  private boolean[][] visited;
  private int[] childLengths;
  private int[] counts;
  private int playerCount;
  private int maxDepth;
  private float winUtil;
  private float inversePlayerCount;
  
  // topology: 
  //   true - a reachable cell (an empty square)
  //   false - an unreachable cell (a solid square)
  public void reset(
      final boolean[][] topology, 
      final int playerCount,
      final int maxDepth) {
    
    this.topology = topology;
    this.playerCount = playerCount;    
    this.maxDepth = maxDepth;
    final int stackSize = maxDepth + playerCount + 1;
    
    rows = topology.length;
    columns = topology[0].length;    
    
    counts = new int[playerCount];
    coordinates = new Coordinate[rows][columns];
    moves = new Coordinate[rows][columns][2][];
    alphas = new float[stackSize][playerCount];
    visited = new boolean[rows][columns];
    
    final List<Coordinate> clonesList = new ArrayList<>();
    final List<Coordinate> skipsList = new ArrayList<>();
    for(int y = rows - 1; y >= 0; y--) {
      for(int x = columns - 1; x >= 0; x--) {
        coordinates[y][x] = new Coordinate(x, y);
        clonesList.clear();
        skipsList.clear();
        if (topology[y][x]) {
          for(int i = 2; i >= -2; i--) {
            for(int j = 2; j >= -2; j--) {

              if (i == 0 && j == 0) {
                continue;
              }

              final int X = x + j;
              final int Y = y + i;
              if (X < 0 || Y < 0 || X >= columns || Y >= rows 
                  || !topology[Y][X]) {
                continue;
              }

              if (abs(i) == 2 || abs(j) == 2) {
                skipsList.add(new Coordinate(X, Y));
              } else {
                clonesList.add(new Coordinate(X, Y));
              }
            }
          }
        }
        clonesList.toArray(moves[y][x][CLONES] 
            = new Coordinate[clonesList.size()]);
        skipsList.toArray(moves[y][x][SKIPS] 
            = new Coordinate[skipsList.size()]);               
        inversePlayerCount = -1f / (playerCount - 1);
      }
    }
        
    final int maxChildren = (rows * columns) * (rows * columns) + 32;
    winUtil = 1000f * maxChildren;
    childLengths = new int[stackSize];    
    children = new Node[stackSize][maxChildren];
    for(int i = stackSize - 1; i >= 0; i--) {
      for(int j = maxChildren - 1; j >= 0; j--) {
        final Node child = children[i][j] = new Node();
        child.playfield = new int[rows][columns];
        child.utilities = new float[playerCount];
      }
    }
  }
  
  public Node hypermax(final Node node, final int depth, final boolean bonus) {
    
    if ((node.source != Coordinate.PASS && depth >= maxDepth)
        || node.nodeType != NodeType.Intermediate) {
      return node;
    }
    
    final int player = node.player;
    final float[] alpha = alphas[depth];
    if (depth == 0) {
      for(int i = playerCount - 1; i >= 0; i--) {
        alpha[i] = -Float.MAX_VALUE;
      }
    } else {
      System.arraycopy(alphas[depth - 1], 0, alphas[depth], 0, playerCount);
    }
    
    final int nextDepth = depth + 1;    
    final Node[] nodes = createChildren(node, depth, bonus);
    Node bestNode = null;
    for(int i = childLengths[depth] - 1; i >= 0; i--) {
      final Node n = hypermax(nodes[i], nextDepth, false);
      if (bestNode == null) {
        bestNode = nodes[i];
        System.arraycopy(n.utilities, 0, bestNode.utilities, 0, playerCount);
      }
      if (n.utilities[player] > alpha[player]) {
        alpha[player] = n.utilities[player];
        bestNode = nodes[i];
        System.arraycopy(n.utilities, 0, bestNode.utilities, 0, playerCount);
      }
      float sum = 0f;
      for(int j = playerCount - 1; j >= 0; j--) {
        sum += alpha[j];
      }
      if (sum >= 0f) {
        break;
      }
    }
    
    return bestNode;
  }

  public Coordinate[][][][] getMoves() {
    return moves;
  }
  
  private Node[] createChildren(final Node node, final int depth, 
      final boolean bonusChildren) {
    
    final int[][] playfield = node.playfield;
    final int player = node.player;
    final int nextPlayer = (node.player + 1 == playerCount) 
        ? 0 : node.player + 1;
    
    for(int y = rows - 1; y >= 0; y--) {
      final boolean[] row = visited[y];
      for(int x = columns - 1; x >= 0; x--) {
        row[x] = false;
      }
    }    
    
    int childLength = 0;
    final Node[] nodes = children[depth];
    if (bonusChildren) {      
      // 3 cherries (skip to anywhere)
      for(int sy = rows - 1; sy >= 0; sy--) {
        for(int sx = columns - 1; sx >= 0; sx--) {
          if (playfield[sy][sx] == player) {
            for(int dy = rows - 1; dy >= 0; dy--) {
              for(int dx = columns - 1; dx >= 0; dx--) {
                if ((dy != sy || dx != sx) && topology[dy][dx] 
                    && playfield[dy][dx] < 0) {
                  final Node n = nodes[childLength++];
                  n.player = nextPlayer;
                  n.setPlayfield(playfield);                  
                  n.playfield[sy][sx] = -1; // cloning never happens
                  n.playfield[dy][dx] = player;
                  n.source = coordinates[sy][sx];
                  n.destination = coordinates[dy][dx];
                  n.bonus = Cherries;
                  final Coordinate[] cs = moves[dy][dx][CLONES];
                  for(int k = cs.length - 1; k >= 0; k--){
                    final Coordinate c = cs[k];
                    final int p = n.playfield[c.y][c.x];
                    if (p >= 0 && p != player) {
                      n.playfield[c.y][c.x] = player;
                    }
                  }
                  evaluate(n, depth);
                }
              }
            }
          }  
        }
      }
        
      // 3 Arcadia Systems logos (swap pieces with an opponent's pieces)
      for(int opponent = playerCount - 1; opponent >= 0; opponent--) {
        if (opponent != player) {
          Coordinate destination = null;
          for(int y = rows - 1; y >= 0; y--) {
            for(int x = columns - 1; x >= 0; x--) {
              if (playfield[y][x] == opponent) {
                destination = coordinates[y][x];
                break;
              }
            }
          }
          if (destination != null) {
            final Node n = nodes[childLength++];
            n.player = nextPlayer;
            n.setPlayfield(playfield);
            n.source = null; // only need to click on an opponent's piece
            n.destination = destination;
            n.bonus = Arcadias;
            for(int y = rows - 1; y >= 0; y--) {
              for(int x = columns - 1; x >= 0; x--) {
                final int cell = n.playfield[y][x];
                if (cell == player) {
                  n.playfield[y][x] = opponent;
                } else if (cell == opponent) {
                  n.playfield[y][x] = player;
                }
              }
            }
            evaluate(n, depth);
          }
        }
      }
      
      // 3 Spots (extra turn, but it only deals with clones due to 3 cherries)
      for(int y = rows - 1; y >= 0; y--) {
        for(int x = columns - 1; x >= 0; x--) {        
          if (playfield[y][x] == player) {
            final Coordinate[] clones = moves[y][x][CLONES];
            for(int i = clones.length - 1; i >= 0; i--) {
              final Coordinate m = clones[i];
              if (playfield[m.y][m.x] < 0) {
                if (visited[m.y][m.x]) {
                  continue;
                } else {
                  visited[m.y][m.x] = true;
                }
                final Node n = nodes[childLength++];
                n.player = nextPlayer;
                n.setPlayfield(playfield);
                n.playfield[m.y][m.x] = player;
                n.source = coordinates[y][x];
                n.destination = m;
                n.bonus = Spots;
                final Coordinate[] cs = moves[m.y][m.x][CLONES];
                for(int k = cs.length - 1; k >= 0; k--){
                  final Coordinate c = cs[k];
                  final int p = n.playfield[c.y][c.x];
                  if (p >= 0 && p != player) {
                    n.playfield[c.y][c.x] = player;
                  }
                }
                evaluate(n, depth);
              }
            }
          } 
        }
      }
      
    } else {
      for(int y = rows - 1; y >= 0; y--) {
        for(int x = columns - 1; x >= 0; x--) {        
          if (playfield[y][x] == player) {
            final Coordinate[][] mvs = moves[y][x];
            for(int i = SKIPS; i >= CLONES; i--) {
              final Coordinate[] ms = mvs[i];
              for(int j = ms.length - 1; j >= 0; j--) {
                final Coordinate m = ms[j];
                if (playfield[m.y][m.x] < 0) {
                  if (i == CLONES) {
                    if (visited[m.y][m.x]) {
                      continue;
                    } else {
                      visited[m.y][m.x] = true;
                    }
                  }
                  final Node n = nodes[childLength++];
                  n.player = nextPlayer;
                  n.setPlayfield(playfield);
                  if (i == SKIPS) {
                    n.playfield[y][x] = -1;
                  }
                  n.playfield[m.y][m.x] = player;
                  n.source = coordinates[y][x];
                  n.destination = m;
                  n.bonus = None;
                  final Coordinate[] cs = moves[m.y][m.x][CLONES];
                  for(int k = cs.length - 1; k >= 0; k--){
                    final Coordinate c = cs[k];
                    final int p = n.playfield[c.y][c.x];
                    if (p >= 0 && p != player) {
                      n.playfield[c.y][c.x] = player;
                    }
                  }
                  evaluate(n, depth);
                }
              }
            }
          } 
        }
      }    
    }
      
    if (childLength == 0) {
      // create pass child
      final Node n = nodes[childLength++];
      n.player = nextPlayer;
      n.setPlayfield(playfield);
      n.source = Coordinate.PASS;
      n.bonus = None;
      evaluate(n, depth);
    }
    childLengths[depth] = childLength;
    
    // randomize the child order    
    for(int i = childLength - 1; i > 0; i--) {
      final int j = random.nextInt(i);
      final Node t = nodes[i];
      nodes[i] = nodes[j];
      nodes[j] = t;
    }
    
    // and, finally sort by descending utility (there could be equal utility
    // values, which is why the sequence is randomized prior)
    Arrays.sort(nodes, 0, childLength, 
        (a, b) -> Float.compare(b.utilities[player], a.utilities[player]));
    
    return nodes;
  }
  
  public void evaluate(final Node node) {
    evaluate(node, 0);
  }
  
  private void evaluate(final Node node, final int depth) {  
    
    final float[] utilities = node.utilities;    
    for(int i = playerCount - 1; i >= 0; i--) {
      utilities[i] = 0f;
      counts[i] = 0;
    }
    
    boolean completelyFilled = true;
    final int[][] playfield = node.playfield;
    for(int y = rows - 1; y >= 0; y--) {
      for(int x = columns - 1; x >= 0; x--) {
        if (topology[y][x]) {
          final int player = playfield[y][x];
          if (player >= 0) {            
            
            counts[player]++;

            // Each piece belonging to the player is worth 20.
            
            // Any square adjacent to the piece that protects it from being
            // captured is worth:
            //  10 - if occupied by the player
            //   5 - if occupied by an opponent
            //   2 - if square is outside of the playfield boundaries
            
            // Any square that can be reached via a skip is worth 1.            
            
            final Coordinate[] clones = moves[y][x][CLONES];
            float score = 36 - (clones.length << 1); // 20 + 2 * outside squares
            for(int i = clones.length - 1; i >= 0; i--) {
              final Coordinate c = clones[i];
              if (playfield[c.y][c.x] >= 0) {
                // 10 for each adjacent square occupied by the player
                //  5 for each adjacent square occupied by an opponent
                score += playfield[c.y][c.x] == player ? 10 : 5; 
              }
            }
            
            final Coordinate[] skips = moves[y][x][SKIPS];
            for(int i = skips.length - 1; i >= 0; i--) {
              final Coordinate c = skips[i];
              if (playfield[c.y][c.x] < 0) {
                score++; // 1 for each possible skip
              }
            }
            
            float antiscore = inversePlayerCount * score;            
            for(int i = playerCount - 1; i >= 0; i--) {
              if (i == player) {
                utilities[i] += score;
              } else {
                utilities[i] += antiscore;
              }
            }
          } else {
            completelyFilled = false;
          }
        }
      }
    }
    
    int maxPlayer = 0;
    int maxCount0 = 0;
    for(int i = playerCount - 1; i >= 0; i--) {
      if (counts[i] > maxCount0) {
        maxCount0 = counts[i];
        maxPlayer = i;
      }
    }

    int maxCount1 = 0;
    for(int i = playerCount - 1; i >= 0; i--) {
      if (i != maxPlayer && counts[i] > maxCount1) {
        maxCount1 = counts[i];
      }
    }

    if (completelyFilled) {
      // Every square has been populated.
      if (maxCount0 == maxCount1) {
        // The top two players have the same number of pieces; it's a draw.        
        node.nodeType = NodeType.Draw;
      } else {
        // The top player has the majority; it's a win.
        node.nodeType = NodeType.Win;        
      }
    } else if (maxCount0 > 0 && maxCount1 == 0) {
      // Only one player has pieces remaining.
      node.nodeType = NodeType.Win;
    } else {
      node.nodeType = NodeType.Intermediate;
    }
    
    if (node.nodeType == NodeType.Win) {
      node.winner = maxPlayer;
      final float W = winUtil - depth;
      final float L = inversePlayerCount * W;
      for(int i = playerCount - 1; i >= 0; i--) {
        utilities[i] += (i == maxPlayer) ? W : L;
      }
    }
  }
}